#include <cassert>
#include <dbm/dbm.h>
#include <system/systemparser.h>
#include <system/task.h>
#include <system/system.h>
#include <system/assignment.h>
#include <system/target.h>
#include <search/successor_generator.h>
#include <unordered_map>
#include <list>
#include <vector>
#include <sstream>
#include <cstring>
#include <climits>
#include "simverifix.h"

using namespace std;
using namespace dbm;


DiscretePart * DiscretePartTable::addDiscretePart(DiscretePart * dp)
{
  list<DiscretePart *> & l = hashtbl[dp->hashKey()];
  if ( l.size() == 0 ){
	l.push_back(dp);
	return dp;
  } else {
	for (list<DiscretePart*>::iterator it = l.begin(); it != l.end(); it++ )
	  {
		if ( (*it)->equal(*dp) ){
		  delete(dp);
		  return *it;
		} 
	  }
	l.push_back(dp);
	return dp;
  }
}

DiscretePartTable::~DiscretePartTable()
{
  unordered_map<uint32_t, list<DiscretePart*> >::iterator it;
  for (it = hashtbl.begin(); it != hashtbl.end(); it++){
	list<DiscretePart *> & l = it->second;
	list<DiscretePart *>::iterator lit = l.begin();
	cerr << "Deleting no : " << it->first << endl;
	for(; lit != l.end(); lit ++){
	  delete(*lit);
	}
  }
}

TaskBundle * readTask(const char * ta, const char * query)
{
  Task * t = SystemParser::parse(string(ta), string(query), true);
  TaskBundle * tb = new TaskBundle(t);
  assert(t && tb);
  return tb;
}

void closeTask(TaskBundle * t)
{
  delete(t);
}

int getNrClocks(TaskBundle * t)
{
  return (int) t->getSystem()->getTotalNrClocks() - 1;
}
int getDimension(TaskBundle * t)
{
  return (int) t->getSystem()->getTotalNrClocks();
}

void displaySystem(TaskBundle * t)
{
  t->getSystem()->display(cout);
}

DiscretePart * getInitialState(TaskBundle * t)
{
  State * s = t->getSystem()->createInitial();
  assert(s);
  DiscretePart * dp = new DiscretePart(s->dp);
  delete(s);
  return t->table.addDiscretePart(dp);
}

/** Here each transition is uniquely identified by its pointer */
void getTransitions(TaskBundle * t, DiscretePart * dp,  std::vector<const Transition*>& succs)
{
  t->succgen->generateEdgesFromDp(dp, succs);
}

DiscretePart * applyTransition(TaskBundle * t, DiscretePart * dp, const Transition * trans)
{
  DiscretePart * next = new DiscretePart(*dp);
  t->succgen->applyEdge(trans->edge1, next);
  if (trans->edge2) {
	t->succgen->applyEdge(trans->edge2, next);
  }
  return t->table.addDiscretePart(next);
}

bool checkTask(TaskBundle * tb)
{
  // TODO
  // Check whether
  // Guards are closed and rectangular
  // All syntactic cycles are of progress
  // Check all resets are to 0, no other integer updates to clocks
  return true;
}

bool isUrgentOrCommitted(TaskBundle * tb, DiscretePart * dp ){
  return tb->getSystem()->isUrgentOrCommitted(dp);
}

void getClockNames(TaskBundle * tb, vector< pair<unsigned,string> > & clock_names)
{
  vector<Clock *> & clocks = tb->getSystem()->clocks;
  for(int i = 1; i < clocks.size(); i++){
	clock_names.push_back(pair<unsigned,string>(clocks[i]->id, clocks[i]->name));
  }
  std::vector<Process*> & procs = tb->getSystem()->procs;
  for(int i = 0; i < procs.size(); i++){
	assert(procs[i]);
	vector<Clock *> & clocks = procs[i]->clocks;
	for(int j = 0; j < clocks.size(); j++){
	  clock_names.push_back(pair<unsigned,string>(clocks[j]->id, procs[i]->name + "." + clocks[j]->name));
	}
  }
}

raw_t * dbm2matrix(dbm_t & g){
  raw_t * data = g.getDBM();
  int dim = g.getDimension();
  raw_t * ret = new raw_t[dim*dim];
  for (int i= 0; i < dim; i++){
	for(int j = 0; j < dim ;j++){
	  ret[i*dim + j] = data[i*dim + j];
	}
  }
  return ret;
}

/** Returns in the encoding [i*dim + j] */
int * guardOfTransition(TaskBundle * tb, const Transition * trans)
{
  int dim = getDimension(tb);
  dbm_t g(dim);
  g.setInit();
  //  cout << g << endl;
  if (!trans->edge1->guard->constrain(&g)) {
	/*       std::cerr<<"Guard not satisfed:"<<*(trans->edge1->guard)<<"\n"; */
	return NULL;
  }
  if (trans->edge2 && !trans->edge2->guard->constrain(&g)) {
	/* std::cerr<<"Guard2 not satisfed:"<<*(trans->edge2->guard)<<"\n"; */
	return NULL;
  }
  return dbm2matrix(g);
  //  return dbm2bounds(&g);
}
int * invariantOfDiscretePart(TaskBundle * tb, const DiscretePart * dp)
{
  size_t dim = getDimension(tb);
  dbm_t g(dim);
  g.setInit();
  for (uint32_t i = 0 ; i < dp->nrProcs; i++){
	int lid = dp->getLocationNr(i);
	tb->getSystem()->procs[i]->getLocation(lid)->inv->clockconstraints.constrain(&g);
  }
  return dbm2matrix(g);
}

/*
int32_t ** invariantOfDiscretePart(TaskBundle * tb, const DiscretePart * dp)
{
  size_t dim = getDimension(tb);
  dbm_t g(dim);
  g.setInit();
  for (uint32_t i = 0 ; i < dp->nrProcs; i++){
	int lid = dp->getLocationNr(i);
	tb->getSystem()->procs[i]->getLocation(lid)->inv->clockconstraints.constrain(&g);
  }
  return dbm2bounds(&g);
}
*/

vector<cindex_t> resetOfTransition(TaskBundle *tb, const Transition * trans)
{
  vector<ClockReset*> & resets1 = trans->edge1->effect->resets;
  vector<cindex_t> rv;
  for (int i = 0 ; i < resets1.size(); i++){
	rv.push_back(resets1[i]->lhs->id);
  }
  if ( trans->edge2 ){
	vector<ClockReset*> & resets2 = trans->edge2->effect->resets;
	for (int i = 0; i < resets2.size(); i++){
	  rv.push_back(resets2[i]->lhs->id);
	}
  }
  return rv;
}

bool dpEqual(DiscretePart * dp1, DiscretePart * dp2)
{
  return dp1->equal(*dp2);
}

void printBoundMatrix(int ** bounds, int dim)
{
  for(cindex_t x = 0; x < dim ; x++){
	for(cindex_t y =0 ; y < dim; y++){
	  if ( bounds[x][y] == dbm_INFINITY ){
		cout << " INF ";
	  } else {
		cout << bounds[x][y] << " ";
	  }
	}
	cout << endl;
  }
  cout << endl;
}

char * stringOfDiscretePart(TaskBundle * tb, const DiscretePart * dp)
{
  stringstream ss;
  dp->prettyPrint(ss, tb->getSystem());
  return strdup(ss.str().c_str());
}

char * stringOfTransition(const Transition * t)
{
  stringstream ss;
  t->display(ss);
  return strdup(ss.str().c_str());
}

bool isTarget(TaskBundle * tb, DiscretePart * dp)
{
  return tb->task->target->isSatBy(dp);
}



void _updateLU(TaskBundle * tb, pair<int,int> * lu, const ClockConstraint * cons){
  int clock = cons->lhs->id;
  int constant = cons->rhs;
  switch (cons->comp){
      case Constraint::LT: 
      case Constraint::LE: 
		if (constant > lu[clock].second){
		  lu[clock].second = constant;
		}
		break;
      case Constraint::GE: 
      case Constraint::GT: 
		if (constant > lu[clock].first){
		  lu[clock].first = constant;
		}
		break;
      case Constraint::EQ:
		if (constant > lu[clock].second){
		  lu[clock].second = constant;
		}
		if (constant > lu[clock].first){
		  lu[clock].first = constant;
		}
		break;
  default: break;
  }
}
void getLUBounds(TaskBundle * tb, vector<int32_t> & floors, vector<int32_t> & ceils, DiscretePart * dp){
  tb->norm->getLUBounds(floors,ceils, dp);
}

/*
pair<int,int> * getLUBounds(TaskBundle * tb)
{
  int dim = getDimension(tb);
  pair<int,int> * lu = new pair<int,int>[dim];
  assert(lu);
  for(int x = 0; x < dim; x++){
	lu[x].first = 0;
	lu[x].second = getMinusInfinity();
  }
  lu[0].second = 0; // Clock 0 is the 0 clock.
  vector<Process *> & proc = tb->getSystem()->procs;
  for(int i = 0; i < proc.size(); i++){
	// Do the invariants
	for(int j= 0; j < proc[i]->getNrLocations(); j++){
	  Conjunction<ClockConstraint> & cons =
		proc[i]->locs[j]->inv->clockconstraints;
	  for (int k = 0; k < cons.size(); k++){
		_updateLU(tb, lu, cons[k]);
	  }
	}
	// Do the guards
	vector<Edge*>::iterator it;
	for(it = proc[i]->edges.begin(); it != proc[i]->edges.end(); it++){
	  Conjunction<ClockConstraint> & cons = (*it)->guard->clockconstraints;
	  for (int j = 0; j < cons.size(); j++){
		  _updateLU(tb, lu, cons[j]);
	  }
	}
  }
  assert(lu);
  return lu;
}
*/
