
open Common
open Pdbm
open Batteries
open Printf


module type TIMED_AUTOMATON =
sig
  module Dbm : GENERIC_DBM
  type timed_automaton
  type discrete_state
  type transition
  val clocks : timed_automaton -> ClockContext.t
  val initial_discrete_state : timed_automaton -> discrete_state
  val transitions_from : timed_automaton -> discrete_state -> transition list
  val transition_fields : timed_automaton -> transition ->
						  (discrete_state * Dbm.t * ClockSet.t * discrete_state)
  val guard_of_transition : timed_automaton -> transition -> Dbm.t
  val invariant_of_discrete_state : timed_automaton -> discrete_state -> Dbm.t
  val print_transition : 'b BatIO.output -> timed_automaton -> transition -> unit
  val is_urgent_or_committed : timed_automaton -> discrete_state -> bool
  val print_discrete_state  : 'b BatIO.output -> timed_automaton -> discrete_state -> unit
  val print_timed_automaton : 'b BatIO.output -> timed_automaton -> unit
  val print_extended_state : 'b BatIO.output -> timed_automaton -> (discrete_state * Dbm.t) -> unit 
  val initial_extended_state : timed_automaton -> discrete_state * Dbm.t
  val is_target : timed_automaton -> discrete_state -> bool
  val lu_bounds : timed_automaton -> discrete_state -> int array -> int array -> unit
end

(** FIXME The following signature only differs by the type of module Dbm *)
module type ENLARGED_TIMED_AUTOMATON =
sig
  module Dbm : ENLARGED_DBM 
  type timed_automaton
  type discrete_state
  type transition
  val clocks : timed_automaton -> ClockContext.t
  val transitions_from : timed_automaton -> discrete_state -> transition list
  val transition_fields : timed_automaton -> transition ->
						  (discrete_state * Dbm.t * ClockSet.t * discrete_state)
  val invariant_of_discrete_state : timed_automaton -> discrete_state -> Dbm.t
  val print_transition : 'b BatIO.output -> timed_automaton -> transition -> unit
  val print_discrete_state  : 'b BatIO.output -> timed_automaton -> discrete_state -> unit
  val print_timed_automaton : 'b BatIO.output -> timed_automaton -> unit
  val print_extended_state : 'b BatIO.output -> timed_automaton -> (discrete_state * Dbm.t) -> unit 
  val is_urgent_or_committed : timed_automaton -> discrete_state -> bool
  val is_target : timed_automaton -> discrete_state -> bool
  val initial_extended_state : timed_automaton -> discrete_state * Dbm.t
  val lu_bounds : timed_automaton -> discrete_state -> int array -> int array -> unit
end


(** UPPAAL Timed Automaton *)

module GenericUppaalAutomaton (Dbm : GENERIC_DBM) =
	struct
		type discrete_state = Simverifix.state
		type transition = discrete_state * Simverifix.trans
		type timed_automaton = 
				{ hard_automaton : Simverifix.t;
					clocks : ClockContext.t;
					guards : (transition,Dbm.t) Hashtbl.t;
					resets : (transition, ClockSet.t) Hashtbl.t;
					invariants : (discrete_state,Dbm.t) Hashtbl.t
				}
		let clocks ta = ta.clocks
		let initial_discrete_state ta = Simverifix.getInitialState ta.hard_automaton
		let _free ta = Simverifix.closeTask ta.hard_automaton
		let transitions_from ta source = 
			(** FIXME Remove rev it's for debuggin *)
			List.rev (List.map (fun tr -> (source,tr)) (Simverifix.getTransitions ta.hard_automaton source))
							 
		let invariant_of_discrete_state ta state = 
			(try Hashtbl.find ta.invariants state with
				 Not_found ->
				 let boundar =
					 Simverifix.flatInvariantOfDiscreteState ta.hard_automaton state in
				 let dim = Simverifix.getDimension ta.hard_automaton in
				 let g = Dbm.from_flat boundar dim in
				 Hashtbl.add ta.invariants state g;
				 g
				|_ as e -> raise e
			)

		let reset_of_transition ta edge = 
			(try Hashtbl.find ta.resets edge with
				 Not_found ->
				 let rlist = Simverifix.resetOfTransition ta.hard_automaton (snd edge) in
				 let rset = List.fold_left 
											(fun set -> fun cl -> (ClockSet.add cl set))
											ClockSet.empty rlist 
				 in
				 Hashtbl.add ta.resets edge rset;
				 rset
			 | _ as e -> raise e
			)

		let source_of_transition ta edge = fst edge
		let target_of_transition ta edge =
			let source, trans = edge in 
			Simverifix.applyTransition ta.hard_automaton source trans 	
				   
		let guard_of_transition ta edge = 
			(try Hashtbl.find ta.guards edge with
				 Not_found ->
				 let boundar = Simverifix.flatGuardOfTransition ta.hard_automaton (snd edge) in
				 let dim = Simverifix.getDimension ta.hard_automaton in
				 let g = Dbm.from_flat boundar dim in
				 Hashtbl.add ta.guards edge g;
				 g
				|_ as e -> raise e
			)

		let lu_bounds ta state floors ceils = 
			Simverifix.getLUBounds ta.hard_automaton state floors ceils

		let transition_fields ta edge = 
			(source_of_transition ta edge, guard_of_transition ta edge,
			 reset_of_transition ta edge, target_of_transition ta edge)

		let is_target ta state = Simverifix.isTarget ta.hard_automaton state

		let is_urgent_or_committed ta state =
			Simverifix.isUrgentOrCommitted ta.hard_automaton state

		let from_file tafile qfile = 
			let task = Simverifix.readTask tafile qfile in
			let cnames = Simverifix.getClockNames task in
			let clocks = ClockContext.create() in 
			List.iter (fun (id,name) -> ClockContext.add clocks name id) cnames;
			{
				hard_automaton = task;
				clocks = clocks;
				guards = Hashtbl.create 100;
				resets = Hashtbl.create 100;
				invariants = Hashtbl.create 100;
			}

		let initial_extended_state ta =
			let initial_ds = Simverifix.getInitialState ta.hard_automaton in
			let dim = (ClockContext.nb_clocks (clocks ta)) + 1 in 
			let z = Dbm.create dim in
			(try
					for i = 0 to dim -1 do
						Dbm.constrain z i 0 Dbm.Alg.zero;
						Dbm.constrain z 0 i Dbm.Alg.zero
					done;
					Dbm.posttime z;
					let inv = (invariant_of_discrete_state ta initial_ds) in
					Dbm.intersect z inv
				with EmptyDBM -> Log.fatal "The initial state is empty\n"
 					 | _ as e-> raise e
			);
			(initial_ds, z)
	
		let print_discrete_state chan ta state = 
			fprintf chan "%s" (Simverifix.stringOfState ta.hard_automaton state)

		let print_extended_state chan ta (ds,g) = 
			print_discrete_state chan ta ds;
			Dbm.pretty_print chan ta.clocks g;
			fprintf chan "\n"

		let print_transition chan ta trans = 
			let _,trans = trans in 
			fprintf chan "%s" (Simverifix.stringOfTransition trans)


		let iter ta fds ftrans =
			let passed = Hashtbl.create 1000 in
			let wait = Stack.create () in 
			let rec visit ()  = 
				match Stack.length wait with
					0 -> ()
				| _ -> 
					 let state = Stack.pop wait in
					 if (Hashtbl.mem passed state) then
						 (
							 visit ()
						 )
					 else (
						 fds state;
						 let trans = transitions_from ta state in
						 Hashtbl.add passed state true;
						 List.iter ftrans trans;
						 let fields = List.map (transition_fields ta) trans in
						 let successors = List.filter 
																(fun s -> not (Hashtbl.mem passed s)) 
																(List.map (fun (s,g,r,t) -> t) fields)
						 in
						 List.iter (fun st -> Stack.push st wait) successors;
						 visit ()
					 ) in
			Stack.push (fst(initial_extended_state ta)) wait;
			visit ()

		let print_timed_automaton chan eta =
			iter eta
					 (fun ds -> fprintf chan "State: "; print_discrete_state chan eta ds; 
											fprintf chan "\nInvariant: ";
											Dbm.pretty_print chan (clocks eta) (invariant_of_discrete_state eta ds);
											fprintf chan "\nTransitions (%d):\n" (List.length (transitions_from eta ds))
					 )
					 (fun trans -> 
						let (s,g,r,t) = transition_fields eta trans in
						print_transition chan eta trans; 
						fprintf chan "\n Internal Guard: ";
						Dbm.pretty_print chan (clocks eta) g;
						fprintf chan "\n Reset(";
						ClockSet.iter (fun x -> fprintf chan "%d " x) r;
						fprintf chan ")\n\n";
					 );
			

	end

module UppaalAutomaton = struct
  include GenericUppaalAutomaton(Guard)
  module Dbm = Guard
end

module type SCALING =
sig
	val scale : int
	val enlarge : int
end

(** UppaalAutomaton obtained by scaling the constants.
This is identical to UppaalAutomaton; the only change is in the guards, invariants and the
lu_bounds function. The lu bounds need to be rescaled at each call, since it relies on an external 
library call which is not aware of the scaling.
 *)
module ScaledUppaalAutomaton (Scaling : SCALING) = struct
  include UppaalAutomaton
  let lu_bounds ta state floors ceils = 
		Simverifix.getLUBounds ta.hard_automaton state floors ceils;
		let dim = Array.length floors in 
		let minfty = Simverifix.getMinusInfinity () in 
		for i = 0 to dim - 1 do
			floors.(i) <- if( floors.(i) = minfty ) then minfty
						  else floors.(i) * Scaling.scale - Scaling.enlarge;
			ceils.(i) <- if ( ceils.(i) = minfty) then minfty
						 else 
						  ceils.(i) * Scaling.scale + Scaling.enlarge
		done

	let invariant_of_discrete_state ta state = 
	  (try Hashtbl.find ta.invariants state with
			 Not_found ->
			 let boundar =
				 Simverifix.flatInvariantOfDiscreteState ta.hard_automaton state in		 
			 let dim = Simverifix.getDimension ta.hard_automaton in
			 for i = 1 to dim - 1 do
				 let sb = 
					 Dbm.Alg.add (Dbm.Alg.weak_from_int Scaling.enlarge)
											 (Dbm.Alg.scale boundar.(i * dim + 0) Scaling.scale)
				 in
				 boundar.(i*dim + 0 ) <- sb;
				 let sb = 
					 Dbm.Alg.add (Dbm.Alg.weak_from_int Scaling.enlarge)
											 (Dbm.Alg.scale boundar.(0*dim + i) Scaling.scale)
				 in
				 boundar.(0*dim + i) <- sb;
			 done;
			 for i = 1 to dim - 1 do
				 for j = 1 to dim - 1 do
					 if ( i <> 0 || j <> 0 ) then(
						 boundar.(i*dim + j) <- Dbm.Alg.infty
					 )
				 done;
			 done;
			 let g = Guard.from_flat boundar dim in
			 Dbm.reduce g;
			 Hashtbl.add ta.invariants state g;
			 g
			|_ as e -> raise e
	  )

  let guard_of_transition ta edge = 
		(try Hashtbl.find ta.guards edge with
			 Not_found ->
			 let boundar = Simverifix.flatGuardOfTransition ta.hard_automaton (snd edge) in
			 let dim = Simverifix.getDimension ta.hard_automaton in
			 for i = 1 to dim - 1 do
			   let sb = 
					 Dbm.Alg.add (Dbm.Alg.weak_from_int Scaling.enlarge)
											 (Dbm.Alg.scale boundar.(i * dim + 0) Scaling.scale)
				 in
				 boundar.(i*dim + 0 ) <- sb;
				 let sb = 
					 Dbm.Alg.add (Dbm.Alg.weak_from_int Scaling.enlarge)
											 (Dbm.Alg.scale boundar.(0*dim + i) Scaling.scale)
				 in
				 boundar.(0*dim + i) <- sb;
			 done;
			 for i = 1 to dim - 1 do
				 for j = 1 to dim - 1 do
					 if ( i <> 0 || j <> 0 ) then(
						 boundar.(i*dim + j) <- Dbm.Alg.infty
					 )
				 done;
			 done;
			 let g = Guard.from_flat boundar dim in
			 Hashtbl.add ta.guards edge g;
			 g
			|_ as e -> raise e
		)

  let transition_fields ta edge = 
		(source_of_transition ta edge, guard_of_transition ta edge,
		 reset_of_transition ta edge, target_of_transition ta edge)

		let initial_extended_state ta =
			let initial_ds = Simverifix.getInitialState ta.hard_automaton in
			let dim = (ClockContext.nb_clocks (clocks ta)) + 1 in 
			let z = Dbm.create dim in
			(try
					for i = 0 to dim -1 do
						Dbm.constrain z i 0 Dbm.Alg.zero;
						Dbm.constrain z 0 i Dbm.Alg.zero
					done;
					Dbm.posttime z;
					let inv = (invariant_of_discrete_state ta initial_ds) in
					Dbm.intersect z inv
				with EmptyDBM -> Log.fatal "The initial state is empty\n"
 					 | _ as e-> raise e
			);
			(initial_ds, z)
	
		let print_discrete_state chan ta state = 
			fprintf chan "%s" (Simverifix.stringOfState ta.hard_automaton state)

		let print_extended_state chan ta (ds,g) = 
			print_discrete_state chan ta ds;
			(*	fprintf chan "\n%s\n" (Dbm.to_string g);	*)
			Dbm.pretty_print chan ta.clocks g;
			fprintf chan "\n"

		let print_transition chan ta trans = 
			let _,trans = trans in 
			fprintf chan "%s" (Simverifix.stringOfTransition trans)

		let print_timed_automaton chan eta =
			iter eta
					 (fun ds -> fprintf chan "State: "; print_discrete_state chan eta ds; 
											fprintf chan "\nInvariant: ";
											Dbm.pretty_print chan (clocks eta) (invariant_of_discrete_state eta ds);
											fprintf chan "\nTransitions (%d):\n" (List.length (transitions_from eta ds))
					 )
					 (fun trans -> 
						let (s,g,r,t) = transition_fields eta trans in
						print_transition chan eta trans; 
						fprintf chan "\n Internal Guard: ";
						Dbm.pretty_print chan (clocks eta) g;
						fprintf chan "\n Reset(";
						ClockSet.iter (fun x -> fprintf chan "%d " x) r;
						fprintf chan ")\n\n";
					 );
			


end



(** EnlargedUppaalAutomaton represents timed automata with 
IEDBMs as its guards, where the guards and invariants are enlarged *)
module EnlargedUppaalAutomaton =
  struct
		include GenericUppaalAutomaton(InfEnlargedDBM)
		module Dbm = InfEnlargedDBM

		let invariant_of_discrete_state ta state = 
			(try Hashtbl.find ta.invariants state with
				 Not_found ->
				 let boundar =
					 Simverifix.flatInvariantOfDiscreteState ta.hard_automaton state in		 
				 let dim = Simverifix.getDimension ta.hard_automaton in
				 let g = Dbm.enlarge_guard(Guard.from_flat boundar dim) in
				 Hashtbl.add ta.invariants state g;
				 g
				|_ as e -> raise e
			)

		let guard_of_transition ta edge = 
			(try Hashtbl.find ta.guards edge with
				 Not_found ->
				 let guardar = Simverifix.flatGuardOfTransition ta.hard_automaton (snd edge) in
				 let dim = Simverifix.getDimension ta.hard_automaton in
				 let g = Dbm.enlarge_guard(Guard.from_flat guardar dim) in
				 Hashtbl.add ta.guards edge g;
				 g
				|_ as e -> raise e
			)

		let transition_fields ta edge = 
			(source_of_transition ta edge, guard_of_transition ta edge,
			 reset_of_transition ta edge, target_of_transition ta edge)

		let from_timed_automaton ta = 
			let guards = Hashtbl.create 100 in
			let invariants = Hashtbl.create 100 in
			{ hard_automaton = ta.UppaalAutomaton.hard_automaton;
				clocks = ta.UppaalAutomaton.clocks;
				guards = guards;
				resets = ta.UppaalAutomaton.resets;
				invariants = invariants
			}
				
		let initial_extended_state ta =
			let initial_ds = Simverifix.getInitialState ta.hard_automaton in
			let dim = (ClockContext.nb_clocks (clocks ta)) + 1 in 
			let z = Dbm.create dim in
			(try
					for i = 0 to dim -1 do
						Dbm.constrain z i 0 Dbm.Alg.zero;
						Dbm.constrain z 0 i Dbm.Alg.zero
					done;
					Dbm.posttime z;
					Dbm.intersect z (invariant_of_discrete_state ta initial_ds);
				with EmptyDBM -> Log.fatal "The initial state is empty\n"
 					 | _ as e-> raise e
			);
			(initial_ds, z)

		let lu_bounds ta state floors ceils = 
			Simverifix.getLUBounds ta.hard_automaton state floors ceils
  end

