open Pdbm
open Common
open AbstractTimedAutomaton
open RobustReach
open ExactReach
open Batteries
open Printf

let usage () = 
  printf 
"symrob 0.1 r2167 
Usage: symrob [options] <system> [query]

Options:
  -e\t\t\t\tExact model checking
  -r\t\t\t\tRobust model checking
  --enlarge N\t\t\tExact model checking applied on syntactic enlargement by N(see also --scale)
  --scale N\t\t\tMultiply all constants by N and run exact model checking
  --initial-width N\t\tSet the enlargement for acceleration threshold to N (default 10)
  --greedy{1,2,3}\t\tGreedy level for the acceleration of the cycles of a prefix (default 3)
  --dfs\t\t\t\tDFS search mode
  --bfs\t\t\t\tBFS search mode

"

let execute_uppaal tafile qfile =
  Log.info (sprintf "Reading timed automaton %s and query %s\n" tafile qfile); flush stdout;
  match !Options.mode with 
  | Options.RobustReachability ->	 
   Log.infof "Running robust reachability (Greedy level: %d)\n" !Options.gammaGreedyLevel;
	 let ta = UppaalAutomaton.from_file tafile qfile in
	 Log.infof "Read TA with %d clocks\n" (ClockContext.nb_clocks (UppaalAutomaton.clocks ta));
	 let eta = EnlargedUppaalAutomaton.from_timed_automaton ta in
	 let reach = match !Options.search with
				 Options.Dfs -> QuantEnlargedUDfs.reach
			 | Options.Bfs ->	QuantEnlargedUBfs.reach 
		 in
		 (match reach eta with
			| SearchTree.Unreachable -> 
				 Log.infof "VERDICT: SAFE for delta < %s\n" (Num.string_of_num !InfEnlargedDBMAlgebra.largest_delta0);
				 Log.SearchStatistics.print_statistics stdout !Options.mode;
			| SearchTree.Reachable(path) ->
				 QuantEnlargedUBfs.print_path stdout eta path;
				 Log.info "VERDICT: REACHABLE\n";
				 Log.SearchStatistics.print_statistics stdout !Options.mode
		 )
  | Options.ExactReachability ->
   Log.info "Running exact reachability\n";
	 let ta = UppaalAutomaton.from_file tafile qfile in
	 Log.infof "Read TA with %d clocks\n" (ClockContext.nb_clocks (UppaalAutomaton.clocks ta));	 
		 let verdict = match !Options.rescaling with
			 | false ->
					let reach = match !Options.search with
							Options.Dfs -> ExactReach.UDfs.reach
						| Options.Bfs ->	ExactReach.UBfs.reach 
					in reach ta
			 | true ->
				Log.info (sprintf "Rescaling (x%d) and enlarging (+%d) timed automaton\n" !Options.scale !Options.enlarge);
				let module Scaling = (struct let scale = !Options.scale let enlarge = !Options.enlarge end) in
				let module ScaledTA = ScaledUppaalAutomaton(Scaling) in
				let module WaitList = (val (match !Options.search with
											  Options.Dfs -> (module ZoneStack : ZONECONTAINER)
											 |Options.Bfs -> (module ZoneQueue : ZONECONTAINER)
										   ) : ZONECONTAINER )
				in
				let module ScaledReach = ExactReachability(WaitList)(ScaledTA) in
				ScaledReach.reach ta
		 in
		 printf "\n";
		 (match verdict with 
			 | SearchTree.Unreachable -> 
					Log.info "VERDICT: SAFE\n";
			 | SearchTree.Reachable(path) ->
					ExactReach.UBfs.print_path stdout ta path;
					Log.info "VERDICT: REACHABLE\n"
		 );
		 Log.SearchStatistics.print_statistics stdout !Options.mode

let main () =
	let isfile name = 
		(String.exists name ".xml") ||	(String.exists name ".q" )
	in
  let args = Array.sub Sys.argv 1 ((Array.length Sys.argv)-1) in
  let options_args = Array.filter (fun s -> not (isfile s)) args in
  let file_args = Array.filter isfile args in
  let prev = ref "" in
  for i = 0 to (Array.length options_args) - 1 do
	match options_args.(i) with
	| "--dfs" -> Options.search := Options.Dfs
	| "--bfs" -> Options.search := Options.Bfs
	| "--enlarged" -> Options.mode := Options.ExactReachability; Options.rescaling := true
	| "--debug" -> Log.set_level Log.Debug
	| "--greedy1" -> Options.gammaGreedyLevel := 1
	| "--greedy2" -> Options.gammaGreedyLevel := 2
	| "--greedy3" -> Options.gammaGreedyLevel := 3
	| "--scale" | "--enlarge" ->
				   Options.mode := Options.ExactReachability; 
				   Options.rescaling := true;
				   prev := options_args.(i)
	| "--initial-width" 
	| "--enlarge-precision" -> prev := options_args.(i)
	| other -> 
	   if ( other.[0] = '-' ) then
		 (String.iter (function
						| '-' -> ()
						| 'q' -> Options.quantified := true
						| 'e' -> Options.mode := Options.ExactReachability 
						| 'r' -> Options.mode := Options.RobustReachability
						| c -> 			 
						   Log.warningf "Unrecognized option: %c\n" c
					  ) other
		 ) 
	   else(
		 try 
		   (match !prev with
			  "--scale" -> Options.scale := int_of_string other
			| "--enlarge" -> Options.enlarge := int_of_string other
			| "--initial-width" -> Options.initial_width := int_of_string other
			| "--enlarge-precision" -> Options.enlarge_precision := Num.num_of_string other
			| _ as s -> usage(); Log.fatalf "Command line parsing problem: %s\n" s
		   );
		   prev := ""
		 with _ -> Log.fatalf "Number expected instead of \"%s\"" other
	   )
	done;
	(** Detect file type *)
  let tafile = ref "" in
  let qfile = ref "" in
  Array.iter (fun filename -> 
			  if ( String.ends_with filename ".xml" ) then
				(
				  tafile := filename;
				)
			  else if (String.ends_with filename ".q") then
				(
				  qfile := filename;
				) 
			  else 
				(
				  Log.fatalf "File extension not recognized: %s\n" filename
				)
			 ) file_args ;
  if ( !tafile = "" ) then
	(
	  usage();
	  Log.fatal "No input files\n"
	);
  if ( !qfile = "" ) then
	(	  
	  qfile := (Filename.chop_suffix !tafile ".xml") ^ ".q";
	  Log.warningf "Guessing query file: %s\n" !qfile
	);
  execute_uppaal !tafile !qfile
					  

let _ = main ()

